# Generated by Django 3.0.7 on 2021-03-30 16:59
from __future__ import annotations

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0008_profile"),
    ]

    operations = [
        migrations.AddField(
            model_name="dataset",
            name="is_default_version",
            field=models.BooleanField(default=False, verbose_name="Default version"),
        ),
        migrations.AddField(
            model_name="dataset",
            name="version",
            field=models.CharField(
                blank=True, max_length=250, null=True, verbose_name="Schema Version"
            ),
        ),
        migrations.AlterField(
            model_name="datasettable",
            name="db_table",
            field=models.CharField(max_length=100),
        ),
    ]



"""Module for SQLAlchemy-based database table creation."""

from __future__ import annotations

import hashlib
import logging
import numbers
from collections import defaultdict
from decimal import Decimal
from typing import cast

from geoalchemy2.types import Geometry
from psycopg import sql
from sqlalchemy import JSON, BigInteger, Boolean, Date, DateTime, Float, Numeric, String, Time
from sqlalchemy.sql.schema import Column, Index, MetaData, Table
from sqlalchemy.types import ARRAY

from schematools import (
    DATABASE_SCHEMA_NAME_DEFAULT,
    MAX_TABLE_NAME_LENGTH,
    SRID_3D,
    TABLE_INDEX_POSTFIX,
    TMP_TABLE_POSTFIX,
)
from schematools.naming import to_snake_case
from schematools.types import DatasetFieldSchema, DatasetSchema, DatasetTableSchema

logger = logging.getLogger(__name__)

FORMAT_MODELS_LOOKUP = {
    "date": Date,
    "time": Time,
    "date-time": DateTime,
    "uri": String,
    "email": String,
    "json": JSON,
}

JSON_TYPE_TO_PG = {
    "string": String,
    "object": String,
    "boolean": Boolean,
    "integer": BigInteger,
    "integer/autoincrement": BigInteger,
    "number": Float,
    "array": ARRAY(String),
    "https://schemas.data.amsterdam.nl/schema@v1.1.0#/definitions/id": String,
    "https://schemas.data.amsterdam.nl/schema@v1.1.0#/definitions/class": String,
    "https://schemas.data.amsterdam.nl/schema@v1.1.0#/definitions/dataset": String,
    "https://schemas.data.amsterdam.nl/schema@v1.1.0#/definitions/schema": String,
    "https://schemas.data.amsterdam.nl/schema@v1.1.1#/definitions/id": String,
    "https://schemas.data.amsterdam.nl/schema@v1.1.1#/definitions/class": String,
    "https://schemas.data.amsterdam.nl/schema@v1.1.1#/definitions/dataset": String,
    "https://schemas.data.amsterdam.nl/schema@v1.1.1#/definitions/schema": String,
}

GEOJSON_TYPE_TO_WKT = {
    "https://geojson.org/schema/Geometry.json": "GEOMETRY",
    "https://geojson.org/schema/Point.json": "POINT",
    "https://geojson.org/schema/Polygon.json": "POLYGON",
    "https://geojson.org/schema/MultiPolygon.json": "MULTIPOLYGON",
    "https://geojson.org/schema/MultiPoint.json": "MULTIPOINT",
    "https://geojson.org/schema/LineString.json": "LINESTRING",
    "https://geojson.org/schema/MultiLineString.json": "MULTILINESTRING",
}


def tables_factory(
    dataset: DatasetSchema,
    metadata: MetaData | None = None,
    db_table_names: dict[str, str | None] | None = None,
    db_schema_names: dict[str, str | None] | None = None,
    limit_tables_to: set | None = None,
    is_versioned_dataset: bool = False,
) -> dict[str, Table]:
    """Generate the SQLAlchemy Table objects base on a `DatasetSchema` definition.

    Args:
        dataset: The Amsterdam Schema definition of the dataset
        metadata: SQLAlchemy schema metadata that groups all tables to a single connection.
        db_table_names: Optional sql table names, keyed on dataset_table_id.
            If not give, db_table_names are inferred from the schema name.
        db_schema_names: Optional database schema names, keyed on dataset_table_id.
            If not given, schema names default to `public`.
        limit_tables_to: Only process the indicated tables (based on table.id).
        is_versioned_dataset: Indicate whether the tables should be created in a private DB
            schema with a version in their name. See also:
            :attr:`.BaseImporter.is_versioned_dataset`. The private
            schema name will be derived from the dataset ID, unless overridden by the
            ``db_schema_name`` parameter.

    The returned tables are keyed on the name of the dataset and table.
    SA Table objects are also created for the junction tables that are needed for relations.

    The nested and through tables that are generated on-the-fly are also taken into account.
    Special care is needed to add postfixes to tables names if the parent table of these
    tables has a non-default name. In that case, these on-the-fly tables also need to
    get an extra postfix in their db name.

    One caveat: The assumption now is that special "overridden" names for tables
    only are used because postfixes are added, and for no other reasons.
    """
    tables = {}
    metadata = metadata or MetaData()
    dataset_tables = dataset.get_tables(include_nested=True, include_through=True)

    # For compatibility and consistency, allow snake-cased table ID's
    db_table_names = _snake_keys(db_table_names) if db_table_names else {}
    db_schema_names = _snake_keys(db_schema_names) if db_schema_names else {}

    # Validate 'limit_tables_to' to avoid unexpected KeyError for the insert_statements
    # later on during the import, as that is based on the tables being generated here.
    if limit_tables_to is not None:
        table_ids = {table.id for table in dataset_tables}
        invalid = set(limit_tables_to) - table_ids - {to_snake_case(id) for id in table_ids}
        if invalid:
            raise ValueError(
                f"limit_tables_to has invalid entries: {', '.join(sorted(invalid))}. "
                f"Available are: {', '.join(sorted(table_ids))}"
            )
        # For compatibility and consistency, the check works on snake-cased identifiers.
        # This is needed because get_table_by_id() also still accepts those values.
        # Otherwise, generate_db_objects(table_id=.., limit_tables_to=..) allows snake-cased
        # identifiers for the first parameter, but needs exact-cased ids for the last parameter.
        limit_tables_to = {to_snake_case(id) for id in limit_tables_to}

    for dataset_table in dataset_tables:
        snaked_table_id = to_snake_case(dataset_table.id)
        snaked_parent_table_id = (
            to_snake_case(dataset_table.parent_table.id)
            if dataset_table.is_nested_table or dataset_table.is_through_table
            else None
        )

        # The junction table for relations are imported separately nowadays.
        # However, nested tables are implemented using a sub-table
        # during import of the main table.
        # So, those nested tables need an SA table object to be able
        # to create the table where data has to be imported into.
        table_object_needed = (
            limit_tables_to is None
            or snaked_table_id in limit_tables_to
            or (dataset_table.is_nested_table and snaked_parent_table_id in limit_tables_to)
        )

        if not table_object_needed:
            logger.debug("tables_factory() - skipping table %s", dataset_table.qualified_id)
            continue

        db_table_description = dataset_table.description
        if (db_table_name := db_table_names.get(snaked_table_id)) is None:
            # No predefined table name, generate a custom one.
            # When the parent has a _new postfix, make sure nested tables also receive that.
            has_postfix = (
                db_schema_names
                and snaked_parent_table_id
                and (parent_table_name := db_table_names.get(snaked_parent_table_id)) is not None
                and parent_table_name.endswith(TMP_TABLE_POSTFIX)
            )
            postfix = TMP_TABLE_POSTFIX if has_postfix else ""
            if is_versioned_dataset:
                db_table_name = dataset_table.db_name_variant(
                    with_dataset_prefix=False, with_version=True, postfix=postfix
                )
            else:
                db_table_name = dataset_table.db_name_variant(postfix=postfix)

        # If schema is None, default to Public. Leave it to None will have a risk that
        # the DB schema that is currently in use will be used to create the table in
        # leading to unwanted/unexepected results
        if (db_schema_name := db_schema_names.get(snaked_table_id)) is None:
            if is_versioned_dataset:
                db_schema_name = to_snake_case(dataset.id)
            else:
                db_schema_name = DATABASE_SCHEMA_NAME_DEFAULT

        columns = [_column_factory(field) for field in dataset_table.get_db_fields()]

        alchemy_table = Table(
            db_table_name,
            metadata,
            *columns,
            comment=db_table_description,
            schema=db_schema_name,
            extend_existing=True,
        )
        alchemy_table.dataset_table = dataset_table
        tables[dataset_table.id] = alchemy_table

    return tables


def _column_factory(field: DatasetFieldSchema) -> Column:
    """Generate a SQLAlchemy column for a single field"""
    try:
        col_type = _get_col_type(field)
    except KeyError:
        raise NotImplementedError(
            f"Field '{field.qualified_id}' type '{field.type}' is not implemented."
        ) from None

    col_kwargs = {"nullable": not field.required}
    if field.is_primary:
        col_kwargs["primary_key"] = True
        col_kwargs["nullable"] = False
        col_kwargs["autoincrement"] = field.type.endswith("autoincrement")

    return Column(field.db_name, col_type, comment=field.description, **col_kwargs)


def _get_col_type(field: DatasetFieldSchema):
    if (field_format := field.format) is not None:
        return FORMAT_MODELS_LOOKUP[field_format]

    # TODO: format takes precedence over multipleof
    # if there is an use case that both can apply for a field definition
    # then logic must be changed
    field_multiple = field.multipleof
    if field_multiple is not None:
        return _numeric_datatype_scale(scale_=field_multiple)

    if field.is_geo:
        is_3d = field.srid in SRID_3D
        return Geometry(
            geometry_type=GEOJSON_TYPE_TO_WKT[field.type],
            srid=field.srid,
            dimension=3 if is_3d else 2,
            spatial_index=False,  # Done manually below for control over index naming
            # use_N_D_index=field.srid in SRID_3D,
        )

    return JSON_TYPE_TO_PG[field.type]


def _numeric_datatype_scale(scale_=None):
    """detect scale from decimal for database datatype scale definition"""
    if (
        isinstance(scale_, numbers.Number)
        and str(scale_).count("1") == 1
        and str(scale_).endswith("1")
    ):
        # TODO: make it possible to set percision too
        # now it defaults to max of 12
        get_scale = Decimal(str(scale_)).as_tuple().exponent
        if get_scale < 0:
            get_scale = get_scale * -1
        return Numeric(precision=12, scale=get_scale)
    else:
        return Numeric


def _snake_keys(value: dict[str, str]) -> dict[str, str]:
    return {to_snake_case(table_id): value for table_id, value in value.items()}


def views_factory(dataset: DatasetSchema, tables: dict[str, Table]) -> dict[str, sql.SQL]:
    """Create VIEW statements.

    The views these statements define are there to provide the illusion that the tables they
    wrap are:

    - not versioned
    - do live in the "public" schema

    This illusion is needed for backwards compatability reasons.

    Args:
        dataset: The dataset currently being processed
        tables: The tables as generated by :func:`tables_factory`

    Returns: A dict mapping table names to VIEW statements.
    """
    dataset_tables: dict[str, DatasetTableSchema] = {
        dst.id: dst for dst in dataset.get_tables(include_nested=True, include_through=True)
    }
    if set(dataset_tables) != set(tables):
        raise ValueError(
            f"mismatch: dataset_tables has {set(dataset_tables)}, tables has {set(tables)}"
        )

    CREATE_VIEW = sql.SQL(
        """
        CREATE OR REPLACE VIEW "public".{view_name} AS
        SELECT *
          FROM {src_schema}.{src_table};
        """
    )
    views: dict[str, sql.SQL] = {
        tn: CREATE_VIEW.format(
            view_name=sql.Identifier(dataset_tables[tn].db_name),
            src_schema=sql.Identifier(tables[tn].schema),
            src_table=sql.Identifier(tables[tn].name),
        )
        for tn in dataset_tables
    }
    return views


def index_factory(
    dataset_table: DatasetTableSchema,
    metadata: MetaData | None = None,
    db_table_name: str | None = None,
    db_schema_name: str | None = None,
    is_versioned_dataset: bool = False,
) -> dict[str, list[Index]]:
    """Generates one or more SQLAlchemy Index objects to work with the JSON Schema.

    Args:
        dataset_table: The Amsterdam Schema definition of the table
        metadata: SQLAlchemy schema metadata that groups all tables to a single connection.
        db_table_name: Optional table name, which is otherwise inferred from the schema name.
        db_schema_name: Optional database schema name, which is otherwise None and
            defaults to "public"
        is_versioned_dataset: Indicate whether the indices should be created in a private DB
            schema with a version in their name. See also:
            :attr:`.BaseImporter.is_versioned_dataset`. The private
            schema name will be derived from the dataset ID, unless overridden by the
            ``db_schema_name`` parameter.

    The returned Index objects are grouped by table names.
    """
    indexes: defaultdict[str, list[Index]] = defaultdict(list)
    _metadata = cast(MetaData, metadata or MetaData())

    if is_versioned_dataset:
        if db_schema_name is None:
            # private DB schema instead of `public`
            db_schema_name = dataset_table.dataset.id
        if db_table_name is None:
            db_table_name = dataset_table.db_name_variant(
                # No dataset prefix as the tables will be created in their own
                # private schema.
                with_dataset_prefix=False,
                with_version=True,
            )
    else:
        if db_schema_name is None:
            db_schema_name = DATABASE_SCHEMA_NAME_DEFAULT
        if db_table_name is None:
            db_table_name = dataset_table.db_name

    table_id = f"{db_schema_name}.{db_table_name}"
    try:
        table: Table = _metadata.tables[table_id]
    except KeyError:
        # When tables_factory() is called with limit_tables_to, it's possible that this
        # main table is not created, and only the M2m are created in a separate import call.
        logger.warning("Table '%s' not found...skipping index creation", table_id)
    else:
        indexes[db_table_name] = [
            _build_identifier_index(table, dataset_table, db_table_name),
            *_build_fk_indexes(table, dataset_table, db_table_name),
            *_build_geo_indexes(table, dataset_table, db_table_name),
            *build_temporal_indexes(table, dataset_table, db_table_name),
        ]

    m2m_indexes = _build_m2m_indexes(metadata, dataset_table, is_versioned_dataset, db_schema_name)
    for table_db_name, through_indexes in m2m_indexes.items():
        indexes[table_db_name].extend(through_indexes)

    return dict(indexes)


def _build_identifier_index(
    table_object: Table, dataset_table: DatasetTableSchema, db_table_name: str
) -> Index:
    """Creates index based on the 'identifier' specification in the Amsterdam schema."""
    identifier_columns: list[Column] = [
        table_object.c[identifier_field.db_name]
        for identifier_field in dataset_table.identifier_fields
    ]

    index_name = _format_index_name(f"{db_table_name}_identifier{TABLE_INDEX_POSTFIX}")
    return Index(index_name, *identifier_columns)


def _build_fk_indexes(
    table_object: Table, dataset_table: DatasetTableSchema, db_table_name: str
) -> list[Index]:
    """Creates an index on Foreign Keys."""
    return [
        Index(
            _format_index_name(f"{db_table_name}_{field.db_name}{TABLE_INDEX_POSTFIX}"),
            table_object.c[field.db_name],
        )
        for field in dataset_table.get_fields(include_subfields=True)
        if field.relation
    ]


def _build_geo_indexes(
    table_object: Table, dataset_table: DatasetTableSchema, db_table_name: str
) -> list[Index]:
    """Creates an index on geometry fields."""
    return [
        Index(
            _format_index_name(f"{db_table_name}_{field.db_name}{TABLE_INDEX_POSTFIX}"),
            table_object.c[field.db_name],
            # Taken from geoalchemy2:
            postgresql_using="gist",
            postgresql_ops=(
                {field.db_name: "gist_geometry_ops_nd"} if field.srid in SRID_3D else {}
            ),
        )
        for field in dataset_table.fields
        if field.is_geo
    ]


def build_temporal_indexes(
    table_object: Table, dataset_table: DatasetTableSchema, db_table_name: str
) -> list[Index]:
    """Creates an index on temporal fields."""
    if dataset_table.temporal:
        fields = set()
        for field in dataset_table._temporal_range_field_ids:
            try:
                fields.add(table_object.c[to_snake_case(field)])
            except KeyError:
                logger.warning("Field '%s' not found...skipping temporal index creation", field)
                return []

        combined_index = Index(
            _format_index_name(f"{db_table_name}_temporal{TABLE_INDEX_POSTFIX}"), *fields
        )

        fields_index = [
            Index(
                _format_index_name(f"{db_table_name}_{field.db_name}{TABLE_INDEX_POSTFIX}"),
                table_object.c[field.db_name],
            )
            for field in dataset_table.temporal.temporal_fields
        ]

        return [combined_index, *fields_index]
    else:
        return []


def _build_m2m_indexes(
    metadata: MetaData,
    dataset_table: DatasetTableSchema,
    is_versioned_dataset: bool,
    db_schema_name: str,
) -> dict[str, list[Index]]:
    """Creates index(es) on the many-to-many tables.

    Those are based on 'relation' specification in the Amsterdam schema.
    """
    indexes: dict[str, list[Index]] = {}
    for field in dataset_table.fields:
        if not field.is_through_table:
            continue

        table = field.through_table

        # create the Index objects
        if is_versioned_dataset:
            table_db_name = table.db_name_variant(with_dataset_prefix=False, with_version=True)
        else:
            table_db_name = table.db_name

        table_id = f"{db_schema_name}.{table_db_name}"
        try:
            table_object = metadata.tables[table_id]
        except KeyError:
            # When tables_factory() is called with limit_tables_to, it's possible that this M2M
            # table is not created yet, and will be created with a separate import call.
            logger.warning("Table '%s' not found...skipping M2M index creation", table_id)
            continue

        indexes[table_db_name] = [
            Index(
                _format_index_name(f"{table_id}_{through_field.db_name}{TABLE_INDEX_POSTFIX}"),
                table_object.c[through_field.db_name],
            )
            for through_field in table.through_fields
        ]
    return indexes


def _format_index_name(index_name: str) -> str:
    """Create a hash value for index_name.

    Postgres DB holds currently 63 max characters for object names.
    To prevent exceeds and collisions, the index names are shortened
    based upon a hash.
    With the blake2s algorithm a digest size is set to 20 bytes,
    which produces a 40 character long hexadecimal string plus
    the additional 4 character postfix of '_idx' (TABLE_INDEX_POSTFIX).
    """
    if len(index_name) <= MAX_TABLE_NAME_LENGTH:
        return index_name
    else:
        return (
            hashlib.blake2s(index_name.encode(), digest_size=20).hexdigest() + TABLE_INDEX_POSTFIX
        )



# Generated by Django 3.0.7 on 2020-09-01 13:00
from __future__ import annotations

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0006_remote_datasets"),
    ]

    operations = [
        migrations.AddField(
            model_name="datasettable",
            name="is_temporal",
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 4.2.20 on 2025-03-31 11:55
from __future__ import annotations

from django.db import migrations


def _remove_remote_datasets(apps, schema_editor):
    Dataset = apps.get_model("datasets", "Dataset")
    remote_datasets = Dataset.objects.filter(endpoint_url__isnull=False).exclude(endpoint_url="")
    for dataset in remote_datasets:
        print(f"- Deleting {dataset.name}")
        dataset.delete()


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0019_datasettable_enable_export"),
    ]

    operations = [
        migrations.RunPython(
            _remove_remote_datasets,
            migrations.RunPython.noop,
        ),
        migrations.RemoveField(
            model_name="dataset",
            name="endpoint_url",
        ),
    ]



# Generated by Django 3.1.12 on 2021-07-20 12:12
from __future__ import annotations

from django.db import migrations, models

import schematools.contrib.django.validators


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0012_schema_data_as_textfield"),
    ]

    operations = [
        migrations.AlterField(
            model_name="profile",
            name="schema_data",
            field=models.TextField(
                validators=[schematools.contrib.django.validators.validate_json],
                verbose_name="Amsterdam Schema Contents",
            ),
        ),
    ]



from __future__ import annotations

import re
from collections import defaultdict
from collections.abc import Iterable

from django.core.management import BaseCommand, CommandError
from django.db import DatabaseError, connection, router, transaction

from schematools.contrib.django.factories import DjangoModelFactory
from schematools.contrib.django.models import Dataset


class Command(BaseCommand):
    help = "Create the tables based on the uploaded Amsterdam schema's."
    requires_system_checks = []  # don't test URLs (which create models)

    def add_arguments(self, parser):
        parser.add_argument(
            "--skip",
            dest="skip",
            nargs="*",
            help="Schemas that need to be skipped",
        )

    def handle(self, *args, **options):
        skip = options.get("skip")
        create_tables(self, Dataset.objects.db_enabled(), allow_unmanaged=True, skip=skip)


def create_tables(
    command: BaseCommand,
    datasets: Iterable[Dataset],
    allow_unmanaged: bool = False,
    base_app_name: str | None = None,
    skip: list[str] | None = None,
    dry_run: bool = False,
) -> None:  # noqa: C901
    """Create tables for all updated datasets.
    This is a separate function to allow easy reuse.
    """
    errors = 0
    command.stdout.write("Creating tables")

    # First create all models. This allows Django to resolve  model relations.
    models = []
    to_be_skipped = set(skip if skip is not None else [])

    # Because datasets are related, we need to 'prewarm'
    # the datasets cache (the DatasetSchema.loader)
    # by accessing the `Dataset.schema` attribute.
    for dataset in datasets:
        dataset.schema  # noqa: B018

    for dataset in datasets:
        if not dataset.enable_db or dataset.name in to_be_skipped:
            continue  # in case create_tables() is called by import_schemas

        factory = DjangoModelFactory(dataset)
        models.extend(factory.build_models())

    # Grouping multiple versions of same model by table name
    models_by_table = defaultdict(list)
    for model in models:
        models_by_table[model._meta.db_table].append(model)

    # Create all tables
    with connection.schema_editor() as schema_editor:
        for db_table_name, models_group in models_by_table.items():
            # Only create tables if migration is allowed
            # - router allows it (not some external database)
            # - table is not a view
            # - model is managed (not by default)
            # - user overrides this (e.g. developer)
            # - create table for latest version of this dataset group
            model = max(models_group, key=lambda model: model._table_schema.version)

            is_view = False

            for view_table in model._dataset.schema.tables:
                if view_table.is_view:
                    command.stdout.write(f"  Skipping view: {db_table_name}")
                    is_view = True

            if is_view:
                continue

            router_allows = router.allow_migrate_model(model._meta.app_label, model)
            if not router_allows:
                command.stdout.write(f"  Skipping externally managed table: {db_table_name}")
                continue

            if not allow_unmanaged and not model._meta.can_migrate(connection):
                command.stderr.write(f"  Skipping non-managed model: {model._meta.db_table}")
                continue

            if dry_run:
                command.stdout.write(
                    f"* Would create table {model._meta.db_table} if it doesn't exist"
                )
            else:
                try:
                    with transaction.atomic():
                        schema_editor.create_model(model)
                except (DatabaseError, ValueError) as e:
                    command.stderr.write(f"  Cannot create table {model._meta.db_table}: {e}")
                    if not re.search(r'relation "[^"]+" already exists', str(e)):
                        errors += 1
                else:
                    command.stdout.write(f"* Created table {model._meta.db_table}")

    if errors:
        raise CommandError("Not all tables could be created")



from __future__ import annotations

import sqlite3

import orjson
import pytest
from sqlalchemy import text

from schematools.exports.csv import export_csvs
from schematools.exports.geojson import export_geojsons
from schematools.exports.geopackage import export_geopackages
from schematools.exports.jsonlines import export_jsonls
from schematools.importer.ndjson import NDJSONImporter


def _load_meetbouten_content(here, engine, meetbouten_schema):
    ndjson_path = here / "files" / "data" / "meetbouten.ndjson"
    importer = NDJSONImporter(meetbouten_schema, engine)
    importer.generate_db_objects("meetbouten", truncate=True, ind_extra_index=False)
    importer.load_file(ndjson_path)


def test_csv_export(here, engine, meetbouten_schema, dbsession, tmp_path):
    """Prove that csv export contains the correct content."""
    _load_meetbouten_content(here, engine, meetbouten_schema)
    with engine.begin() as connection:
        export_csvs(connection, meetbouten_schema, str(tmp_path), [], [], 1)
        with open(tmp_path / "meetbouten_meetbouten.csv") as out_file:
            assert out_file.read() == (
                "Identificatie,Ligtinbuurtid,Merkcode,Merkomschrijving,Geometrie\n"
                "1,10180001.1,12,De meetbout,SRID=28992;POINT(119434 487091.6)\n"
            )


def test_csv_export_only_actual(here, engine, ggwgebieden_schema, dbsession, tmp_path):
    """Prove that csv export contains only the actual records, not the history."""
    ndjson_path = here / "files" / "data" / "ggwgebieden-history.ndjson"
    importer = NDJSONImporter(ggwgebieden_schema, engine)
    importer.generate_db_objects("ggwgebieden", truncate=True, ind_extra_index=False)
    importer.load_file(ndjson_path)
    with engine.begin() as connection:
        export_csvs(connection, ggwgebieden_schema, str(tmp_path), [], [], 1)
        with open(tmp_path / "ggwgebieden_ggwgebieden.csv") as out_file:
            lines = out_file.readlines()
            assert len(lines) == 2  # includes the headerline
            assert lines[1].split(",")[0] == "2"  # volgnummer == 2


def test_jsonlines_export(here, engine, meetbouten_schema, dbsession, tmp_path):
    """Prove that jsonlines export contains the correct content."""
    with engine.begin() as connection:
        _load_meetbouten_content(here, engine, meetbouten_schema)
        export_jsonls(connection, meetbouten_schema, str(tmp_path), [], [], 1)
        with open(tmp_path / "meetbouten_meetbouten.jsonl") as out_file:
            result = orjson.loads(out_file.read())
            result["geometrie"]["coordinates"][1] = round(result["geometrie"]["coordinates"][1], 5)
            result["geometrie"]["coordinates"][0] = round(result["geometrie"]["coordinates"][0], 5)
            assert result == {
                "identificatie": 1,
                "ligtInBuurtId": "10180001.1",
                "merkCode": "12",
                "merkOmschrijving": "De meetbout",
                "geometrie": {"type": "Point", "coordinates": [4.86497, 52.37055]},
            }


# We have to skip this test, ogr2og2 is not available on github
# We need to think of a flag to enable this test locally with `skipif`.
@pytest.mark.skip()
def test_geopackage_export(here, engine, meetbouten_schema, dbsession, tmp_path):
    """Prove that geopackage export contains the correct content."""
    _load_meetbouten_content(here, engine, meetbouten_schema)
    with engine.begin() as connection:
        export_geopackages(connection, meetbouten_schema, str(tmp_path), [], [])
    sqlite3_conn = sqlite3.connect(tmp_path / "meetbouten_meetbouten.gpkg")
    cursor = sqlite3_conn.cursor()
    cursor.execute(text("select * from rtree_sql_statement_geometrie"))
    res = cursor.fetchall()
    assert res == [(1, 119434.0, 119434.0, 487091.59375, 487091.65625)]
    cursor.execute(
        """
            select identificatie, ligt_in_buurt_id, merk_code, merk_omschrijving from sql_statement
       """
    )
    res = cursor.fetchall()
    assert res == [(1, "10180001.1", "12", "De meetbout")]


def test_geojson_export(here, engine, meetbouten_schema, dbsession, tmp_path):
    """Prove that geojson export contains the correct content."""
    with engine.begin() as connection:
        _load_meetbouten_content(here, engine, meetbouten_schema)
        export_geojsons(connection, meetbouten_schema, str(tmp_path), [], [], 1)
        with open(tmp_path / "meetbouten_meetbouten.geojson") as out_file:
            result = orjson.loads(out_file.read())
            feature = result["features"][0]
            feature["geometry"]["coordinates"][1] = round(feature["geometry"]["coordinates"][1], 5)
            feature["geometry"]["coordinates"][0] = round(feature["geometry"]["coordinates"][0], 5)
            assert feature == {
                "type": "Feature",
                "properties": {
                    "identificatie": 1,
                    "ligtInBuurtId": "10180001.1",
                    "merkCode": "12",
                    "merkOmschrijving": "De meetbout",
                },
                "geometry": {"type": "Point", "coordinates": [4.86497, 52.37055]},
            }



from __future__ import annotations

from datetime import date
from typing import Final

from sqlalchemy import Boolean, text
from sqlalchemy.sql.ddl import CreateSchema
from sqlalchemy.sql.elements import TextClause

from schematools.importer.base import BaseImporter
from schematools.importer.ndjson import NDJSONImporter

SCHEMA_EXISTS: Final[TextClause] = text(
    """
    SELECT EXISTS(SELECT schema_name
              FROM information_schema.schemata
              WHERE schema_name = :schema_name) AS exists
    """
).columns(exists=Boolean)
TABLE_EXISTS: Final[TextClause] = text(
    """
    SELECT EXISTS(SELECT table_name
              FROM information_schema.tables
              WHERE table_schema = :schema_name
                AND table_name = :table_name
                AND table_type = 'BASE TABLE') AS exists;
    """
).columns(exists=Boolean)
VIEW_EXISTS: Final[TextClause] = text(
    """
    SELECT EXISTS(SELECT table_name
              FROM information_schema.tables
              WHERE table_schema = :schema_name
                AND table_name = :view_name
                AND table_type = 'VIEW') AS exists;
    """
).columns(exists=Boolean)


def test_camelcased_names_during_import(here, engine, bouwblokken_schema, dbsession):
    ndjson_path = here / "files" / "data" / "gebieden.ndjson"
    importer = NDJSONImporter(bouwblokken_schema, engine)
    importer.generate_db_objects("bouwblokken", truncate=True, ind_extra_index=False)
    last_record = importer.load_file(ndjson_path)
    assert dict(last_record) == {
        "begin_geldigheid": "2008-03-12",
        "eind_geldigheid": "2010-10-19",
        "id": 2,
        "ligt_in_buurt_id": 34,
    }
    assert last_record.source == {
        "beginGeldigheid": "2008-03-12",
        "eindgeldigheid": "2010-10-19",
        "id": 2,
        "ligtinbuurt": 34,
        "schema": "irrelevant",
    }
    with engine.begin() as conn:
        records = (
            conn.execute(text("SELECT * FROM bouwblokken_bouwblokken_v1 ORDER BY id"))
            .mappings()
            .all()
        )
    assert len(records) == 2
    assert set(records[0].keys()) == {
        "id",
        "begin_geldigheid",
        "eind_geldigheid",
        "ligt_in_buurt_id",
    }
    assert records[0]["begin_geldigheid"] == date(2006, 1, 12)
    assert records[0]["eind_geldigheid"] == date(2008, 11, 14)


def test_skip_duplicate_keys_in_batch_during_import(here, engine, bouwblokken_schema, dbsession):
    """Prove that the ndjson, which has a duplicate record, does not lead to an exception"""
    ndjson_path = here / "files" / "data" / "gebieden-duplicate-id.ndjson"
    importer = NDJSONImporter(bouwblokken_schema, engine)
    importer.generate_db_objects("bouwblokken", truncate=True, ind_extra_index=False)
    last_record = importer.load_file(ndjson_path)
    with engine.begin() as conn:
        records = (
            conn.execute(text("SELECT * FROM bouwblokken_bouwblokken_v1 ORDER BY id"))
            .mappings()
            .all()
        )
    assert records == [
        # Only one inserted, and no crash happened.
        {
            "begin_geldigheid": date(2006, 1, 12),
            "eind_geldigheid": date(2008, 11, 14),
            "id": "1",
            "ligt_in_buurt_id": "12",
        }
    ]

    # Despite not being inserted, the record should still be returned as last one,
    # because this the last one processed (useful for cursor tracking).
    assert last_record == {
        "begin_geldigheid": "2008-03-12",
        "eind_geldigheid": "2010-10-19",
        "id": 1,
        "ligt_in_buurt_id": 34,
    }


def test_skip_duplicate_keys_in_db_during_import_with_existing_value(
    here, engine, bouwblokken_schema, dbsession
):
    """Prove that the ndjson, which has a duplicate record, does not lead to an exception"""
    ndjson_path = here / "files" / "data" / "gebieden.ndjson"
    importer = NDJSONImporter(bouwblokken_schema, engine)
    importer.generate_db_objects("bouwblokken", truncate=True, ind_extra_index=False)
    importer.load_file(ndjson_path)


def test_skip_duplicate_keys_in_db_during_import_with_duplicate_in_next_batch(
    here, engine, bouwblokken_schema, dbsession
):
    """Prove that the ndjson, which has a duplicate record, does not lead to an exception"""
    ndjson_path = here / "files" / "data" / "gebieden.ndjson"
    importer = NDJSONImporter(bouwblokken_schema, engine)
    importer.generate_db_objects("bouwblokken", truncate=True, ind_extra_index=False)
    importer.load_file(ndjson_path)
    importer.load_file(ndjson_path)


def test_numeric_datatype_scale(
    here, engine, woningbouwplannen_schema, gebieden_schema, dbsession
):
    """Prove that when multipleOf is used in schema,
    it's value is used to set the scale of the numeric datatype"""
    importer = BaseImporter(woningbouwplannen_schema, engine)
    importer.generate_db_objects("woningbouwplan", ind_tables=True, ind_extra_index=True)
    with engine.begin() as conn:
        records = (
            conn.execute(
                text(
                    """
                SELECT data_type, numeric_scale
                    FROM information_schema.columns
                    WHERE table_schema = 'public'
                      AND table_name = 'woningbouwplannen_woningbouwplan_v1'
                      AND column_name = 'avarage_sales_price';
                """
                )
            )
            .mappings()
            .all()
        )
    assert records[0]["data_type"] == "numeric"
    assert records[0]["numeric_scale"] == 4


def test_invalid_numeric_datatype_scale(
    here, engine, woningbouwplannen_schema, gebieden_schema, dbsession
):
    """Prove that when invaldi multipleOf i.e. 0.000 is used in schema,
    the datatype is in that case just plain numeric without scale"""
    importer = BaseImporter(woningbouwplannen_schema, engine)
    importer.generate_db_objects("woningbouwplan", ind_tables=True, ind_extra_index=True)
    with engine.begin() as conn:
        record = (
            conn.execute(
                text(
                    """
                SELECT data_type, numeric_scale
                    FROM information_schema.columns
                    WHERE table_schema = 'public'
                        AND table_name = 'woningbouwplannen_woningbouwplan_v1'
                        AND column_name = 'avarage_sales_price_incorrect'
                       OR column_name = 'avarage_sales_price_incorrect_zero'
                """
                )
            )
            .mappings()
            .all()
        )
    assert record[0]["data_type"] == "numeric"
    assert not record[0]["numeric_scale"]
    assert record[1]["data_type"] == "numeric"
    assert not record[1]["numeric_scale"]


def test_biginteger_datatype(here, engine, woningbouwplannen_schema, gebieden_schema, dbsession):
    """Prove that when biginter is used as datatype in schema, the datatype
    in the database is set to datatype int(8) instead of int(4)"""
    importer = BaseImporter(woningbouwplannen_schema, engine)
    importer.generate_db_objects("woningbouwplan", ind_tables=True, ind_extra_index=True)
    with engine.begin() as conn:
        results = conn.execute(
            text(
                """
            SELECT data_type, numeric_precision
                FROM information_schema.columns
                WHERE table_schema = 'public'
                  AND table_name = 'woningbouwplannen_woningbouwplan_v1'
                  AND column_name = 'id';
            """
            )
        )
        record = results.fetchone()
    assert record.data_type == "bigint"
    assert record.numeric_precision == 64


def test_add_column_comment(here, engine, woningbouwplannen_schema, dbsession):
    """Prove that a column comment is added as defined in the schema as field description"""
    importer = BaseImporter(woningbouwplannen_schema, engine)
    importer.generate_db_objects("woningbouwplan", ind_tables=True, ind_extra_index=False)
    with engine.begin() as conn:
        results = conn.execute(
            text(
                """
            SELECT pgd.description
                FROM pg_catalog.pg_statio_all_tables AS st
                         INNER JOIN pg_catalog.pg_description pgd ON (pgd.objoid = st.relid)
                         INNER JOIN information_schema.columns c ON (pgd.objsubid = c.ordinal_position
                    AND c.table_schema = st.schemaname AND c.table_name = st.relname)
                WHERE table_name = 'woningbouwplannen_woningbouwplan_v1'
                  AND column_name = 'projectnaam';
            """
            )
        )
        record = results.fetchone()
    assert record.description == "Naam van het project"


def test_add_table_comment(here, engine, woningbouwplannen_schema, dbsession):
    """Prove that a table comment is added as defined in the schema as table description"""
    importer = BaseImporter(woningbouwplannen_schema, engine)
    importer.generate_db_objects("woningbouwplan", ind_tables=True, ind_extra_index=False)
    with engine.begin() as conn:
        results = conn.execute(
            text(
                """
            SELECT OBJ_DESCRIPTION('public.woningbouwplannen_woningbouwplan_v1'::REGCLASS) AS description;
            """
            )
        )
        record = results.fetchone()
    assert (
        record.description == "De aantallen vormen de planvoorraad. "
        "Dit zijn niet de aantallen die definitief worden gerealiseerd. "
        "Ervaring leert dat een deel van de planvoorraad wordt opgeschoven. "
        "Niet alle woningbouw initiatieven doorlopen de verschillende plaberumfasen. "
        "Met name kleinere particuliere projecten worden in de regel pas toegevoegd aan "
        "de monitor zodra er een intentieovereenkomst of afsprakenbrief is getekend."
    )


def test_create_table_db_schema(here, engine, woningbouwplannen_schema, dbsession):
    """Prove that a table is created in given DB schema."""
    with engine.begin() as conn:
        conn.execute(CreateSchema("schema_foo_bar", if_not_exists=True))

    importer = BaseImporter(woningbouwplannen_schema, engine)
    importer.generate_db_objects(
        "woningbouwplan", "schema_foo_bar", ind_tables=True, ind_extra_index=False
    )
    with engine.begin() as conn:
        results = conn.execute(
            text(
                """
            SELECT schemaname FROM pg_tables WHERE tablename = 'woningbouwplannen_woningbouwplan_v1'
            """
            )
        )
        record = results.fetchone()
    assert record.schemaname == "schema_foo_bar"


def test_create_table_no_db_schema(here, engine, woningbouwplannen_schema, dbsession):
    """Prove that a table is created in DB schema public if no DB schema is given."""
    importer = BaseImporter(woningbouwplannen_schema, engine)
    importer.generate_db_objects("woningbouwplan", None, ind_tables=True, ind_extra_index=False)

    with engine.begin() as conn:
        results = conn.execute(
            text(
                """
            SELECT schemaname FROM pg_tables WHERE tablename = 'woningbouwplannen_woningbouwplan_v1'
            """
            )
        )
        record = results.fetchone()
    assert record.schemaname == "public"


def test_generate_db_objects_is_versioned_dataset(
    here, engine, woningbouwplannen_schema, dbsession
):
    """Prove that dataset is created in private DB schema with versioned tables."""
    with engine.connect() as conn:
        assert not conn.scalar(SCHEMA_EXISTS, {"schema_name": "woningbouwplannen"})

        importer = BaseImporter(woningbouwplannen_schema, engine)
        importer.generate_db_objects(
            "woningbouwplan", ind_tables=True, ind_extra_index=False, is_versioned_dataset=True
        )
        assert conn.scalar(SCHEMA_EXISTS, {"schema_name": "woningbouwplannen"})
        for table_name in (
            "woningbouwplan_v1",
            "woningbouwplan_buurten_v1",
            "woningbouwplan_buurten_as_scalar_v1",
        ):
            assert conn.scalar(
                TABLE_EXISTS, {"schema_name": "woningbouwplannen", "table_name": table_name}
            )
        for view_name in (
            "woningbouwplannen_woningbouwplan_v1",
            "woningbouwplannen_woningbouwplan_buurten_v1",
            "woningbouwplannen_woningbouwplan_buurten_as_scalar_v1",
        ):
            assert conn.scalar(VIEW_EXISTS, {"schema_name": "public", "view_name": view_name})


def test_create_table_temp_name(engine, db_schema, woningbouwplannen_schema, gebieden_schema):
    """Prove that a table is created in DB schema with the temporary name
    as definied in a dictionary."""
    importer = BaseImporter(woningbouwplannen_schema, engine)
    importer.generate_db_objects(
        "woningbouwplan",
        db_table_name="foo_bar",
        ind_tables=True,
        ind_extra_index=False,
    )

    with engine.begin() as conn:
        results = conn.execute(
            text("SELECT tablename FROM pg_tables WHERE tablename LIKE 'foo_%%'")
        )
        names = {r[0] for r in results.fetchall()}
    assert names == {"foo_bar"}
